from .model import Model

class Genre(Model):
    """Represents a Manga Tag."""
    __slots__ = (
        "id", "name", "description", "url",
        "provider", "instance"
    )

    def __init__(self, provider):
        super().__init__(provider)

        self.id: str = ""
        self.name: str = ""
        self.description: str = ""
        self.url: str = ""
        self.provider = provider
        self.instance = self

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return str(self)



from .model import Model

class Page(Model):
    """Represents a Page"""
    __slots__ = (
        "id",
        "image_link",
        "provider", "instance"
    )

    def __init__(self, provider):
        super().__init__(provider)

        self.id: str = ""
        self.image_link: str = ""

    def __str__(self) -> str:
        return self.image_link

    def __repr__(self) -> str:
        return self.image_link



from pprint import pprint

from mext import Mext

latest_url = 'https://asurascans.com/'
manga_url = 'https://asurascans.com/manga/rise-from-the-rubble/'
chapter_url = 'https://asurascans.com/rise-from-the-rubble-chapter-138/'


class AsuraScansTest:

    def __init__(self) -> None:
        self.mext = Mext()

    def test_latest(self):
        pprint(self.mext.get_latest(latest_url))

    def test_manga(self):
        pprint(self.mext.get_manga(manga_url))

    def test_manga_list(self):
        pprint(self.mext.get_manga_list(latest_url))

    def test_chapter(self):
        pprint(self.mext.get_chapter(chapter_url).to_dict())

    def test_chapter_list(self):
        chapter_list = self.mext.get_manga_chapters(manga_url)
        for chapter in chapter_list:
            print(chapter.number, chapter.name)


if __name__ == '__main__':
    atest = AsuraScansTest()
    atest.test_latest()
    atest.test_manga()
    atest.test_chapter()
    atest.test_chapter_list()



import json
import functools
from textwrap import wrap

from mext import enums


# Get Selenium Status Code

def get_status(logs):
    for log in logs:
        if log['message']:
            d = json.loads(log['message'])
            try:
                content_type = 'text/html' in d['message']['params']['response']['headers']['content-type']
                response_received = d['message']['method'] == 'Network.responseReceived'
                if content_type and response_received:
                    return d['message']['params']['response']['status']
            except:
                pass



from .model import Model


class Group(Model):
    """Represents a MangaDex Group."""
    __slots__ = (
        "id", "name", "desc", "website", "irc_server", "irc_channel", "discord", "email",
        "locked", "official", "verified", "leader", "members", "created_at", "updated_at", "client"
    )

    def __init__(self, provider):
        super().__init__(provider)
        
        self.id: str = ""


        self.provider = provider
    
    def __str__(self) -> str:
        return f"{self.name}"
    
    def __repr__(self) -> str:
        return str(self)


from typing import List
from datetime import datetime

from mext import models

from .model import Model


class Chapter(Model):
    """Represents a Chapter."""
    __slots__ = (
        "id", "name", "number", "volume", "language", "special", "pages", "page_count",
        "manga", "group", "uploader", "url", "created_at", "updated_at",
        "provider", "instance"
    )

    def __init__(self, provider):
        super().__init__(provider)
        
        self.id: str = ""
        self.name: str = ""
        self.number: str = ""
        self.volume: float = float(0)
        self.language: str = ""
        self.special: bool = False
        self.oneshot: bool = False
        self.pages: List[models.Page] = ""
        self.page_count: int = int(0)
        self.manga: models.Manga = None
        self.group: models.Group = None
        self.uploader: str = None
        self.url: str = ""
        self.created_at: datetime = None
        self.updated_at: datetime = None
        self.provider = provider
        self.instance = self

    def __str__(self) -> str:
        srep = f"Chapter {self.number}"
        if self.name:
            srep = f"{srep} - {self.name}"
        return srep

    def __repr__(self) -> str:
        return str(self)



import time
from datetime import datetime
from typing import List, Dict, Union, Type

from mext import models, enums
from mext.models.model import Model

from mext.exceptions import *
from mext.provider import Provider

# Copy from https://github.com/Proxymiity/MangaDex.py

INCLUDE_ALL = ["cover_art", "manga", "chapter", "scanlation_group",
               "author", "artist", "user", "leader", "member"]


class MangadexLinkNames(enums.AttributeEnum):
    Anilist = 'al'
    AnimePlanet = 'ap'
    Bookwalker = 'bw'
    Mangaupdates = 'mu'
    Novelupdates = 'nu'
    Kitsu = 'kt'
    Amazon = 'amz'
    eBookJapan = 'ebj'
    MyAnimeList = 'mal'
    CDJapan = 'cdj'
    Raw = 'raw'
    OfficialEnglishTL = 'engtl'


mangadex_language_map = {
    'zh-hk': 'zh',
    'zh-ro': 'zh',
    'ja-ro': 'ja',
    'ko-ro': 'ko',
    'zh': 'zh',
    'en': 'en',
    'ja': 'ja',
    'ko': 'ko',
}


class MangadexOrg(Provider):

    def __init__(self, *args, **kwargs):
        self.language = "en"
        super(MangadexOrg, self).__init__(*args, **kwargs)

        self.rate_limit = 0.25
        self.api_url = "https://api.mangadex.org"
        self.net_api_url = "https://api.mangadex.network"

    def process_url(self, url):
        super().process_url(url)

    def get_uuid(self, url):
        self.process_url(url)
        return self.parsed_url.path.split('/')[2]

    def populate_manga_data(self, data) -> models.Manga:
        manga = models.Manga(self)

        manga.id = data.get("id")
        manga.url = f"https://{self.siteUrl}/title/{manga.id}"

        _attrs = data.get("attributes")
        _rel = data.get("relationships", [])
        manga.title = _attrs.get("title").get("en")
        manga.alts = [list(alt_dict.values())[0]
                      for alt_dict in _attrs.get("altTitles")]
        manga.description = _attrs.get("description").get("en")
        manga.language = mangadex_language_map[
            _attrs.get("originalLanguage").lower()
        ]
        manga.comic_type = enums.ComicTypesLanguage[
            data.get("type").lower()
        ].name.title()
        # manga.last_chapter = _attrs.get("lastChapter")

        demo_genre = models.Genre(self)
        demo_genre.name = (
            _attrs.get("publicationDemographic") or ""
        ).title()

        manga.status = enums.StatusTypes(
            _attrs.get("status").lower()
        ).name
        manga.year = int(_attrs.get("year") or manga.year)

        manga.created_at = datetime.fromisoformat(_attrs.get("createdAt"))
        manga.updated_at = datetime.fromisoformat(_attrs.get("updatedAt"))

        try:
            links = []

            if _attrs.get("links"):

                for link_key, link_value in _attrs.get("links").items():
                    link = models.Link(self)
                    link.name = MangadexLinkNames(link_key).name
                    link.link = link_value
                    links.append(link)

                manga.links = links
        except Exception as e:
            print("Exception getting Link:\n", e)

        try:
            genres = []

            if _attrs.get("tags"):

                for genre_attr in _attrs.get("tags"):
                    genre = models.Genre(self)
                    genre.name = genre_attr["attributes"]["name"].get("en")

                    if not genre.name:
                        continue

                    genres.append(genre)
                manga.genres = genres
        except Exception as e:
            print("Exception getting Genre:\n", e)

        try:
            authors = []

            _author_list = [
                x["attributes"] for x in _rel if x["type"] == "author"
            ]

            for author_rel in _author_list:
                author = models.Person(self)
                author.name = author_rel["name"]
                authors.append(author)

            manga.authors = authors
        except (IndexError, KeyError) as e:
            print("Exception getting Authors:\n", e)

        try:
            artists = []

            _artist_list = [
                x["attributes"] for x in _rel if x["type"] == "artist"
            ]
            for artist_rel in _artist_list:
                artist = models.Person(self)
                artist.name = artist_rel["name"]
                artists.append(artist)

            manga.artists = artists
        except (IndexError, KeyError) as e:
            print("Exception getting Artists:\n", e)

        try:
            covers = []
            _cover_list = [
                x["attributes"] for x in _rel if x["type"] == "cover_art"
            ]
            
            for cover_rel in _cover_list:
                cover = models.Cover(self)
                cover.volume = float(cover_rel.get("volume") or 0)
                cover.filename = cover_rel.get("fileName")
                cover.image_link = f"https://uploads.mangadex.org/covers/{manga.id}/{cover.filename}"
                covers.append(cover)

            manga.all_covers = covers

            if covers:
                manga.current_cover = covers[-1]

        except (IndexError, KeyError) as e:
            print("Exception getting Covers:\n", e)

        return manga

    def populate_chapter_data(self, data):
        chapter = models.Chapter(self)
        chapter.id = data.get("id")
        chapter.url = f"https://{self.siteUrl}/chapter/{chapter.id}"

        _attrs = data.get("attributes")
        _rel = data.get("relationships", [])
        chapter.volume = float(_attrs.get("volume") or 0)
        chapter.number = (_attrs.get("chapter") or "")
        chapter.page_count = int(_attrs.get("pages"))

        chapter.name = (_attrs.get("title") or "")
        chapter.language = _attrs.get("translatedLanguage").lower()
        chapter.created_at = datetime.fromisoformat(
            _attrs.get("publishAt")
        )
        chapter.updated_at = datetime.fromisoformat(
            _attrs.get("updatedAt")
        )

        return chapter

    def get_latest(self, url: str, params: Dict) -> List[models.Manga]:
        includes = INCLUDE_ALL
        params = params or {}
        if includes:
            params = {"includes[]": includes}
        params = {
            **params,
            "translatedLanguage[]": "en",
            "order[readableAt]": "desc",
            "limit": 100,
        }
        req = self.client.http.get(
            f"{self.api_url}/chapter",
            params=params
        )
        if req.status_code == 200:
            resp = req.json()
            data_list = resp["data"]

            latest_list = []
            for data in data_list:

                _rel = data["relationships"]
                _manga = [
                    x for x in _rel if x["type"] == "manga"
                ][0]

                manga = self.populate_manga_data(_manga)
                latest_list.append(manga)

            return latest_list
        elif req.status_code == 404:
            raise NoContentError(req)
        else:
            raise APIError(req)

        return

    def get_manga(self, url, params) -> models.Manga:

        uuid = self.get_uuid(url)
        includes = INCLUDE_ALL
        params = params or {}
        if includes:
            params = {"includes[]": includes}
        req = self.client.http.get(
            f"{self.api_url}/manga/{uuid}", params=params)
        if req.status_code == 200:
            resp = req.json()
            data = resp["data"]

            manga = self.populate_manga_data(data)
            manga.url = url

            return manga
        elif req.status_code == 404:
            raise NoContentError(req)
        else:
            raise APIError(req)

    def get_chapter(self, url, params) -> models.Chapter:
        uuid = self.get_uuid(url)
        includes = INCLUDE_ALL
        params = params or {}
        if includes:
            params = {"includes[]": includes}
        req = self.client.http.get(
            f"{self.api_url}/chapter/{uuid}",
            params=params
        )
        if req.status_code == 200:
            resp = req.json()
            data = resp["data"]

            chapter = self.populate_chapter_data(data)
            chapter.url = url

            return self.read_chapter(chapter)
        elif req.status_code == 404:
            raise NoContentError(req)
        else:
            raise APIError(req)

    def get_manga_chapters(self, url, params) -> List[models.Chapter]:
        uuid = self.get_uuid(url)
        includes = INCLUDE_ALL
        params = params or {
            "translatedLanguage[]": "en",
            "order[chapter]": "desc",
        }
        if includes:
            params["includes[]"] = includes
        return self._retrieve_chapters(
            f"{self.api_url}/manga/{uuid}/feed",
            models.Chapter,
            call_limit=100,
            params=params
        )

    def get_cover(self, url, params) -> models.Cover:
        uuid = self.get_uuid(url)
        req = self.client.http.get(f"{self.api_url}/cover/{uuid}")
        if req.status_code == 200:
            resp = req.json()
            return models.Cover(resp["data"], self)
        elif req.status_code == 404:
            raise NoContentError(req)
        else:
            raise APIError(req)

    def read_chapter(self, chapter: models.Chapter) -> models.Chapter:
        uuid = chapter.id
        params = {"forcePort443": False}
        req = self.client.http.get(
            f"{self.api_url}/at-home/server/{uuid}", params=params
        )
        if req.status_code == 200:
            data = req.json()

            base_url = data.get("baseUrl")
            _ch = data.get("chapter")
            hash = _ch.get("hash")
            files = _ch.get("data")
            pages = []

            for filename in files:
                page = models.Page(self)
                page.image_link = f"{base_url}/data/{hash}/{filename}"
                pages.append(page)

            chapter.pages = pages

            return chapter
        else:
            raise APIError(req)

    def _retrieve_chapters(self, url: str, chapter: models.Chapter,
                           limit: int = 0, call_limit: int = 500,
                           params: dict = None) -> List[models.Chapter]:
        params = params or {}
        data = []
        offset = 0
        resp = None
        remaining = True
        if "limit" in params:
            params.pop("limit")
        if "offset" in params:
            params.pop("offset")
        while remaining:
            p = {
                "limit": limit if limit <= call_limit and
                limit != 0 else call_limit,
                "offset": offset
            }
            p = {**p, **params}
            req = self.client.http.get(url, params=p)
            if req.status_code == 200:
                resp = req.json()
                data += [x for x in resp["data"]]
            elif req.status_code == 204:
                pass
            else:
                raise APIError(req)
            if limit and len(data) >= limit:
                break
            if resp is not None:
                remaining = resp["total"] > offset + call_limit
                offset += call_limit
            else:
                remaining = False
            if remaining:
                time.sleep(self.rate_limit)
        if not data:
            raise NoResultsError()
        return [self.populate_chapter_data(x) for x in data if int(x["attributes"]["pages"]) > 0]



from enum import Enum
from typing import List, Dict


class BaseEnum(Enum):

    @classmethod
    def __contains__(cls, name) -> bool:
        return name in [c.name for c in cls]

    @classmethod
    def list(cls) -> List:
        return [c.value for c in cls]

    @classmethod
    def keys(cls) -> List:
        return [c.name for c in cls]

    @classmethod
    def dict(cls) -> Dict:
        return {c.name: c.value for c in cls}

    @classmethod
    def reverse_dict(cls) -> Dict:
        return {c.value: c.name for c in cls}


class AttributeEnum(str, BaseEnum):
    pass


class StatusTypes(AttributeEnum):
    Ongoing = 'ongoing'
    Completed = 'completed'
    Hiatus = 'hiatus'
    Dropped = 'dropped'
    ComingSoon = 'comic Soon'


class ComicTypesLanguage(AttributeEnum):
    manga = 'ja'
    manhua = 'zh'
    manhwa = 'ko'
    webtoon = 'en'

class ContentRatingTypes(AttributeEnum):
    """Lowest to Highest order Content Rating"""
    Safe = 'safe'
    Suggestive = 'suggestive'
    Erotica = 'erotica'
    Pornographic = 'pornographic'


from typing import Dict


class Model:

    def __init__(self, provider):
        self.provider_name: str = provider.name

    @property
    def __dict__(self) -> Dict:
        return self.to_dict()

    def __iter__(self):
        for field in self.__slots__:
            yield (field, getattr(self, field))

    def __str__(self) -> str:
        return "{} {}".format(self.provider_name, self.__class__.__name__)

    def to_dict(self) -> Dict:
        return {field: getattr(self, field) for field in self.__slots__}



from .http import Http
from . import selenium as Selenium

from mext import utils

class Client:
    http = None
    selenium = None

    def __init__(self, use_client='http') -> None:
        self.name = use_client

        if not use_client:
            raise Exception(
                "Error initializing client, use_client value not supported : {}".format(use_client)
            )
        elif use_client == 'selenium':
            Selenium.make_driver(browser='chrome')
            self.selenium = Selenium.get_driver()
        else:
            self.http = Http()
    
    @property
    def is_http(self):
        return self.name == 'http'
    
    @property
    def is_selenium(self):
        return self.name == 'selenium'
    
    @property
    def status_code(self):
        status_code = None
        if self.is_selenium:
            logs = self.selenium._driver.get_log('performance')
            status_code = utils.get_status(logs)
        else:
            self.http.status_code



from .model import Model

class Link(Model):
    """Represents a Link."""
    __slots__ = (
        "id", "name", "image_link", "link"
    )
    
    def __init__(self, provider):
        super().__init__(provider)

        self.id: str = ""
        self.name: str = ""
        self.image_link: str = ""
        self.link: str = ""
    
    def __str__(self) -> str:
        return f"{self.name} - {self.link}"
    
    def __repr__(self) -> str:
        return str(self)
    


from mext.bases import MangaStreamBase


class AsuraScansCom(MangaStreamBase):

    def __init__(self, name, siteUrl):
        self.language = 'en'
        super(AsuraScansCom, self).__init__(name, siteUrl)



import re

from datetime import datetime
from urllib.parse import ParseResult, urlencode, urlparse, urlunparse

from bs4 import BeautifulSoup

from mext import enums
from mext import models
from mext.provider import Provider


class MangaStreamBase(Provider):

    def __init__(self, name, siteUrl):
        self.name = name
        super(MangaStreamBase, self).__init__(name, siteUrl)

    def process_chapter_name(self, chapter_text):
        chapter_number, chapter_name = None, ''
        if chapter_text:
            chapter_text = chapter_text.strip()

            match_patterns = [
                r'preview|promotional\s+video',
                r'([0-9]*[.]?[0-9]+)\s*\-?\s*(.*)'
            ]
            if re.search(match_patterns[0], chapter_text, re.IGNORECASE):
                chapter_name = chapter_text
                chapter_number = float(0)

            elif re.search(match_patterns[1], chapter_text):
                rsearch = re.search(match_patterns[1], chapter_text)
                len_rsearch = len(rsearch.groups())

                if len_rsearch > 0 and rsearch.group(1):
                    chapter_number = rsearch.group(1)

                    if len_rsearch > 1 and rsearch.group(2):
                        chapter_name = rsearch.group(2)

            chapter_number = chapter_number \
                if chapter_number not in [None, ''] \
                else chapter_text

            try:
                chapter_number = float(chapter_number)
            except:
                if chapter_text:
                    chapter_number = float(0)
                    chapter_name = chapter_text
        else:
            chapter_number = None

        return chapter_number, chapter_name

    def get_latest(self, url, params):

        parsed_url = urlparse('/'.join(url.split('/')[0:3]))

        pr = ParseResult(
            scheme=parsed_url.scheme,
            netloc=parsed_url.netloc,
            path='/manga/',
            params='',
            query=urlencode({'page': page, 'order': 'update'}),
            fragment=''
        )

        mangalist_url = urlunparse(pr)

        req = self.client.http.get(mangalist_url)
        self.find_error(mangalist_url)

        latest_list = []

        soup = BeautifulSoup(req.content, 'lxml')

        listing_element = soup.select_one('div.mrgn div.listupd')

        if listing_element:

            for update_element in listing_element.select('div.bs'):
                if update_element:
                    manga = models.Manga(self)

                    url_element = update_element.select_one(
                        'div.bsx > a[href][title]')
                    if url_element:
                        manga.url = url_element.attrs['href']

                    cover_element = update_element.select_one(
                        'div.limit > img.ts-post-image[alt][src]')
                    if cover_element:
                        cover = models.Cover(self)
                        cover.image_link = cover_element.attrs['src']
                        manga.all_covers.append(cover)
                        manga.current_cover = manga.all_covers[-1]

                    type_element = update_element.select_one('span.type')
                    type_text = type_element.string if type_element else ''
                    if type_text:
                        manga.comic_type = type_text

                    title_element = update_element.select_one(
                        'div.bigor > div.tt')
                    title_text = title_element.string if title_element else ''
                    if title_text:
                        manga.title = title_text.strip()
                    else:
                        raise Exception(
                            'Could not find comic title for one of the comics')

                    chapter_element = update_element.select_one(
                        'div.bigor div.epxs')
                    chapter_text = chapter_element.string if chapter_element else ''
                    chapter_number = re.findall(
                        'Chapter (\d+)', chapter_text) if chapter_text else ''

                    if chapter_number:
                        latest_chapter = models.Chapter(self)
                        latest_chapter.number = str(float(chapter_number[0]))
                        manga.last_chapter = latest_chapter

                latest_list.append(manga)

        return latest_list

    def get_manga(self, url, params):

        req = self.client.http.get(url)
        self.find_error(url)

        manga = models.Manga(self)
        manga.url = url

        soup = BeautifulSoup(req.content, 'lxml')

        wrong_field_values = ['-', 'N/A']

        # ID
        id_element = soup.select_one('div.rt div.bookmark[data-id]')

        if id_element:
            id_text = id_element.attrs['data-id']

            if id_text:
                manga.id = id_text.strip()

        # Cover
        thumb_element = soup.select_one(
            'div.thumb[itemtype="https://schema.org/ImageObject"] > img[itemprop="image"]'
        )

        if thumb_element:
            thumb_url = thumb_element.attrs.get('src')

            if thumb_url:
                cover = models.Cover(self)
                cover.image_link = thumb_url.strip()
                manga.current_cover = cover
                manga.all_covers.append(cover)

        # Background Image
        bg_element = soup.select_one(
            'div.bigcover div.ime img[src]'
        )

        if bg_element and bg_element.attrs.get('src'):
            manga.banner_picture = bg_element.attrs.get('src').strip()

        # Title
        title_element = soup.select_one('h1.entry-title')

        title_text = title_element.string if title_element else ''
        if title_text:
            trans = title_text.maketrans({
                "‘": "'",
                "’": "'",
                "“": "'",
                "”": "'",
            })
            manga.title = title_text.strip()\
                .translate(trans)
        else:
            raise Exception('Could not get title for comic')

        # Status
        status_element = soup.select('div.tsinfo div.imptdt')[0]

        status_text = status_element.select_one(
            'i').string if status_element else ''
        if status_text:
            manga.status = status_text

        # Type
        type_element = soup.select('div.tsinfo div.imptdt')[1]
        type_text = type_element.select_one('a').string if type_element else ''
        if type_text:
            manga.comic_type = type_text

        # Language
        if type_text:
            type_text = type_text.strip().lower()
            if type_text in enums.ComicTypesLanguage.dict():
                manga.language = enums.ComicTypesLanguage[type_text].value

        # Alternative Titles
        alt_element = soup.select_one('div.infox div.wd-full')

        alt_names = []
        if alt_element and \
                alt_element.select_one('b') and \
                alt_element.select_one('b').string.strip() == 'Alternative Titles':
            alt_text = alt_element.select_one('span').string.strip()

            if alt_text and alt_text not in wrong_field_values:
                alt_delimiters = ['|', ',']
                for delimiter in alt_delimiters:
                    for alt_name in alt_text.split(delimiter):
                        alt_names.append(alt_name.strip())

                    if alt_names:
                        break

        manga.alts = alt_names

        # Description
        description_element = soup.select_one(
            'div.infox div.entry-content.entry-content-single'
        )

        description_text = description_element.text if description_element else ''
        if description_text:
            trans = description_text.maketrans({
                "‘": "'",
                "’": "'",
                "“": "'",
                "”": "'",
            })
            manga.description = description_text.strip()\
                .translate(trans)

        for metadata_element in soup.select('div.infox div.fmed'):

            if metadata_element:
                field_name_element = metadata_element.select_one('b')
                if field_name_element:
                    if field_name_element.string:
                        field_name = field_name_element.string
                    else:
                        field_name = field_name_element.text

                    field_name = field_name.strip()
                else:
                    continue

                field_value_element = metadata_element\
                    .select_one('span')
                if field_value_element:
                    if field_value_element.string:
                        field_value = field_value_element.string.strip()
                    else:
                        field_value = field_value_element.text.strip()

                    field_value = field_value.strip()

                    if field_value in wrong_field_values:
                        continue
                else:
                    continue
            else:
                continue

            # Author
            if field_name == 'Author':
                author_names = []
                author_names_text = field_value

                for name in author_names_text.split('/'):
                    if name:
                        person = models.Person(self)
                        person.name = name.strip()
                        author_names.append(person)

                manga.authors = author_names

            # Artist
            if field_name == 'Artist':
                artist_names = []
                artist_names_text = field_value

                for name in artist_names_text.split('/'):
                    if name:
                        person = models.Person(self)
                        person.name = name.strip()
                        artist_names.append(person)

                manga.artists = artist_names

            # Posted On
            if field_name == 'Posted On':

                posted_on_dt = field_value_element\
                    .select_one('time').attrs.get('datetime')

                if posted_on_dt:
                    manga.created_at = datetime.fromisoformat(posted_on_dt)

            # Updated On
            if field_name == 'Updated On':

                updated_on_dt = field_value_element\
                    .select_one('time').attrs.get('datetime')

                if posted_on_dt:
                    manga.created_at = datetime.fromisoformat(updated_on_dt)

        # Genres
        genre_elements = soup.select('span.mgen > a')

        genres = []
        if genre_elements:
            for genre_element in genre_elements:
                name = genre_element.text.strip()
                if name:
                    genre = models.Genre(self)
                    genre.name = name
                    genres.append(genre)

        manga.genres = genres

        # Extra data

        # Followers
        followers_element = soup.select_one('div.rt div.bmc')

        if followers_element:
            followers_text = followers_element.text.strip()
            followers_find = re.findall(
                'Followed by ([\d]+) people', followers_text)

            if followers_find:
                manga.followers = int(followers_find[0])

        # Rating
        rating_element = soup.select_one(
            'div.rating div[itemprop="ratingValue"]')

        if rating_element:
            rating_text = rating_element.text.strip()

            if rating_text:
                manga.rating = float(rating_text)

        # Return complete Manga data
        return manga

    def get_chapter(self, url, params):

        req = self.client.http.get(url)
        self.find_error(url)

        chapter = models.Chapter(self)
        chapter.url = url

        soup = BeautifulSoup(req.content, 'lxml')

        breadcrumb_element = soup.select_one(
            'ol[itemtype="http://schema.org/BreadcrumbList"]')
        if breadcrumb_element:
            lis = breadcrumb_element.select('li[itemprop="itemListElement"]')

            title_element = lis[1]
            title_text = title_element\
                .select_one('span[itemprop="name"]').string if title_element else ''

            chapter_bc = lis[2]
            chapter_name_text = chapter_bc\
                .select_one('span[itemprop="name"]').string if chapter_bc else ''

            if title_text and chapter_name_text:
                chapter_text = chapter_name_text\
                    .strip().replace(title_text, "", 1).strip()
                chapter_text = re.findall(
                    r'(?:chapter)?\s*(.+)', chapter_text, re.IGNORECASE)

                if chapter_text:
                    chapter_number, chapter_name = self.process_chapter_name(
                        chapter_text[0]
                    )

                    if chapter_number not in [None, '']:
                        chapter.number = str(chapter_number)
                    else:
                        raise Exception(
                            f"Error extracting chapter number from text text {chapter_text}"
                        )

                    if chapter_name:
                        chapter.name = str(chapter_name)
                else:
                    print('Chapter number not found in text {chapter_text}')

        pages = []
        pages_element = soup.select_one('div#readerarea')

        for page_element in pages_element.select('p > img[src][width][height]'):
            page_url = page_element.attrs['src']
            page = models.Page(self)
            page.image_link = page_url
            pages.append(page)

        chapter.pages = pages

        return chapter

    def get_manga_chapters(self, url, params):

        req = self.client.http.get(url)
        self.find_error(url)

        chapter_list = []

        soup = BeautifulSoup(req.content, 'lxml')

        chapters_element = soup.select_one('ul.clstyle')

        for chapter_element in chapters_element.select('li[data-num]'):
            chapter = models.Chapter(self)

            if chapter_element:

                chapter_url_element = chapter_element\
                    .select_one('div.eph-num > a')
                if chapter_url_element:
                    chapter.url = chapter_url_element.attrs['href']

                chapter_text = chapter_element.attrs['data-num']
                chapter_number, chapter_name = self.process_chapter_name(
                    chapter_text
                )

                if chapter_number not in [None, '']:
                    chapter.number = str(chapter_number)
                else:
                    continue

                if chapter_name:
                    chapter.name = str(chapter_name)

                date_element = chapter_element.select_one('span.chapterdate')
                date_text = date_element.string if date_element else ''
                if date_text:
                    chapter.created_at = datetime.strptime(
                        date_text, '%B %d, %Y')

            chapter_list.append(chapter)

        return chapter_list



import re

from datetime import datetime
from urllib.parse import ParseResult, urlencode, urlparse, urlunparse

from bs4 import BeautifulSoup

from mext import enums
from mext import models
from mext.provider import Provider


class ReaperScansCom(Provider):

    def __init__(self, name, siteUrl):
        self.language = 'en'
        super(ReaperScansCom, self).__init__(name, siteUrl)

    def get_latest(self, url: str, page):
        return

    def get_manga(self, url: str, page):

        req = self.client.http.get(url)
        self.find_error(url)

        manga = models.Manga(self)
        manga.url = url

        soup = BeautifulSoup(req.content, 'lxml')

        wrong_field_values = ['-', 'N/A']

        # Description
        description_element = soup.select_one('p[tabindex="0"]')

        description_text = description_element.text if description_element else ''
        if description_text:
            trans = description_text.maketrans({
                "‘": "'",
                "’": "'",
                "“": "'",
                "”": "'",
            })
            manga.description = description_text.strip()\
                .translate(trans)

        other_elements_selector = "body > div.flex.flex-col.h-screen.justify-between > main > div.mx-auto.py-8.grid.max-w-3xl.grid-cols-1.gap-4.sm\\:px-6.lg\\:max-w-screen-2xl.lg\\:grid-flow-col-dense.lg\\:grid-cols-3 > section > div:nth-child(1) > div > div > dl > div"
        other_elements = soup.select(other_elements_selector)

        for metadata_element in other_elements:

            if metadata_element:
                field_name_element = metadata_element.select_one('dt')
                if field_name_element:
                    if field_name_element.string:
                        field_name = field_name_element.string
                    else:
                        field_name = field_name_element.text

                    field_name = field_name.strip()
                else:
                    continue

                field_value_element = metadata_element\
                    .select_one('dd')
                if field_value_element:
                    if field_value_element.string:
                        field_value = field_value_element.string.strip()
                    else:
                        field_value = field_value_element.text.strip()

                    field_value = field_value.strip()

                    if field_value in wrong_field_values:
                        continue
                else:
                    continue
            else:
                continue

            # Status
            if field_name == 'Release Status':
                if field_value:
                    manga.status = enums.StatusTypes('ongoing').name

            # Type
            type_element = soup.select('div.tsinfo div.imptdt')[1]
            type_text = type_element\
                .select_one('a').string if type_element else ''
            if type_text:
                manga.comic_type = type_text

        return manga

    def get_chapter(self, *args, **kwargs):
        return super(ReaperScansCom, self).get_chapter(*args, **kwargs)

    def get_manga_chapters(self, *args, **kwargs):
        return super(ReaperScansCom, self).get_manga_chapters(*args, **kwargs)



import json

class NotYetSupported(Exception):

    def __init__(self, message="This is not yet supported", *args, **kwargs):
        self.message = message
        super().__init__(message, *args, **kwargs)

class APIError(Exception):
    def __init__(self, r):
        self.status = r.status_code
        self.data = None
        try:
            self.data = r.json()
        except json.decoder.JSONDecodeError:
            self.data = r.content

class NoContentError(APIError):
    pass


class LoginError(APIError):
    pass


class NotLoggedInError(Exception):
    pass


class NoResultsError(Exception):
    pass



from pprint import pprint

from mext import Mext

latest_url = 'https://mangadex.org/'
manga_url = 'https://mangadex.org/title/d1c0d3f9-f359-467c-8474-0b2ea8e06f3d/bocchi-sensei-teach-me-mangadex'
chapter_url = 'https://mangadex.org/chapter/e183d3f4-fde0-4288-a1ed-8547490f84b3'


class MangadexTest:

    def __init__(self) -> None:
        self.mext: Mext = Mext()

    def test_latest(self):
        pprint(self.mext.get_latest(latest_url))

    def test_manga(self):
        pprint(self.mext.get_manga(manga_url).to_dict())

    def test_chapter(self):
        pprint(self.mext.get_chapter(chapter_url).to_dict())

    def test_chapter_list(self):
        chapter_list = self.mext.get_manga_chapters(manga_url)
        for chapter in chapter_list:
            print(chapter.number, chapter.name)


if __name__ == '__main__':
    atest = MangadexTest()
    # atest.test_latest()
    atest.test_manga()
    atest.test_chapter()
    atest.test_chapter_list()



import cloudscraper


class Http:

    retry_count = 20
    _headers = {}
    proxies = {}
    referer = ''
    status_code = None
    user_agent = '{} {}'.format(
        'Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:91.0)',
        'Gecko/20100101 Firefox/91.0'
    )
    default_lang = 'ru-RU,ru;q=0.8,en-US;q=0.5,en;q=0.3'
    webp_header = 'text/html,application/xhtml+xml,application/xml;q=1.0,image/webp,image/apng,*/*;q=1.0'

    def __init__(self) -> None:
        self.proxies = {}
        self.cookies = {}
        self._history = []
        self.scraper = cloudscraper.create_scraper()

    def request(self, url: str, headers: dict = {}, data=None, cookies: dict = {}, method: str = 'get', files=None, **kwargs):
        if not isinstance(headers, dict) or headers:
            headers = {}

        headers.setdefault('User-Agent', self.user_agent)
        if self.referer:
            headers.setdefault('Referer', self.referer)
        headers.setdefault('Accept-Language', self.default_lang)
        headers['Accept'] = self.webp_header

        return self.scraper.request(
            url=url,
            proxies=self.proxies,
            headers=headers,
            cookies=cookies,
            method=method,
            files=files,
            data=data,
            **kwargs
        )

    def get(self, url: str, headers: dict = {}, cookies: dict = {}, **kwargs):
        return self.request(
            url=url,
            headers=headers,
            cookies=cookies,
            method='get',
            **kwargs
        )

    def get_stream(self, url: str, headers: dict = {}, cookies: dict = {}, **kwargs):
        kwargs['stream'] = True
        return self.get(url, headers, cookies, **kwargs)

    def reset_proxy(self):
        self.proxies = {}

    def set_proxy(self, proxy):
        self.reset_proxy()
        if isinstance(proxy, dict):
            self.proxies['http'] = proxy.get('http', None)
            self.proxies['https'] = proxy.get('https', None)
        elif isinstance(proxy, str):
            self.proxies['http'] = proxy



from urllib.parse import urlparse
from typing import List, Dict

from mext import models, providers
from mext.provider import Provider


class Mext:

    def __init__(self) -> None:
        self.provider: Provider = None

    def set_provider(self, url):
        parsed_url = urlparse(url)

        provider_instance = providers.get_provider_instance(
            netloc=parsed_url.netloc
        )

        self.provider = provider_instance

    def get_provider(self, url: str) -> Provider:
        if not self.provider:
            self.set_provider(url)

        return self.provider

    def get_latest(self, url: str, params: Dict = {}) -> List[models.Manga]:
        return self.get_provider(url).get_latest(url, params)

    def get_manga(self, url: str, params: Dict = {}) -> models.Manga:
        return self.get_provider(url).get_manga(url, params)

    def get_manga_list(self, url: str, params: Dict = {}) -> List[models.Manga]:
        return self.get_provider(url).get_manga_list(url, params)

    def get_chapter(self, url: str, params: Dict = {}) -> models.Chapter:
        return self.get_provider(url).get_chapter(url, params)

    def get_manga_chapters(self, url: str, params: Dict = {}) -> List[models.Chapter]:
        return self.get_provider(url).get_manga_chapters(url, params)

    @property
    def all_providers(self):
        return providers.providers_json



from urllib.error import HTTPError
from urllib.parse import urlparse
from typing import List, Dict

from mext import client, utils

from mext import models


class Provider:

    def __init__(self, name, siteUrl):
        self.name = name
        self.siteUrl = siteUrl
        self.language = ''

        self.client = client.Client('http')
    
    def __str__(self) -> str:
        return self.name
    
    def __str__(self) -> str:
        return str(self)

    def process_url(self, url):
        self.parsed_url = urlparse(url)
        self.scheme = self.parsed_url.scheme
        self.netloc = self.parsed_url.netloc

    def get_latest(self, url: str, params: Dict = {}) -> List[models.Manga]:
        """Gets list of updated mangas."""
        raise NotImplementedError

    def get_manga(self, url: str, params: Dict = {}) -> models.Manga:
        """Gets a manga with a specific url."""
        raise NotImplementedError

    def get_manga_list(self, url: str, params: Dict = {}) -> List[models.Manga]:
        """Gets a list of Manga."""
        raise NotImplementedError

    def get_chapter(self, url: str, params: Dict = {}) -> models.Chapter:
        """Gets a chapter with a specific url."""
        raise NotImplementedError

    def get_manga_chapters(self, url: str, params: Dict = {}) -> List[models.Chapter]:
        """Gets chapters associated with a specific Manga."""
        raise NotImplementedError

    def get_cover(self, url: str) -> models.Cover:
        """Gets cover data associated with a specific Manga."""

    def find_error(self, url):
        if self.client.is_selenium:
            logs = self.client.selenium._driver.get_log('performance')
            status_code = utils.get_status(logs)
        else:
            status_code = self.client.http.status_code

        http_error_msg = ""

        if isinstance(status_code, int):
            if 400 <= status_code < 500:
                http_error_msg = (
                    f"{status_code} Client Error for url: {url}"
                )

            elif 500 <= status_code < 600:
                http_error_msg = (
                    f"{status_code} Server Error for url: {url}"
                )

        if http_error_msg:
            raise HTTPError(url=url, code=status_code,
                            msg=http_error_msg, hdrs=None, fp=None)



import setuptools

README = open('README.md').read()

setuptools.setup(
    name='mext',
    version='2.0.5',
    url='https://github.com/modbender/manga-extractor',
    description='A simple manga extractor. Extracts any comic info, chapter list and chapter pages. In development.',
    long_description=README,
    long_description_content_type='text/markdown',
    author='Yashas H R',
    author_email='rameshmamathayashas@gmail.com',
    packages=setuptools.find_packages(),
    package_dir={'mext': 'mext'},
    install_requires=[
        'requests',
        'lxml>=4.6.0',
        'cloudscraper',
        'beautifulsoup4>=4.10.0',
    ],
    python_requires='>=3.5',
    platforms=['any'],
    classifiers=[
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Environment :: Console',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Internet :: WWW/HTTP',
    ]
)



import os

from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.common.desired_capabilities import DesiredCapabilities

import undetected_chromedriver as uc


class ChromeDriver:
    arguments = []
    _driver = None

    def __init__(self, arguments: list = []):
        self.arguments = arguments
    
    def _make_driver(self):

        self.capabilities = DesiredCapabilities.CHROME.copy()
        self.capabilities['goog:loggingPrefs'] = {'performance': 'ALL'}

        self.options = uc.ChromeOptions()
        self.options.headless = False

        if self.arguments:
            argument_list = self.arguments
        else:
            argument_list = [
                # '--no-sandbox', # Insecure
                '--mute-audio',
                '--disable-gpu',
                '--no-first-run',
                '--disable-extensions'
                '--no-service-autorun',
                '--password-store=basic',
                # '--disable-dev-shm-usage',
                '--disable-blink-features',
                '--disable-blink-features=AutomationControlled',
            ]
        self.options.arguments.extend(argument_list)

        self._driver = uc.Chrome(
            options=self.options,
            desired_capabilities=self.capabilities
        )
    
    def init_driver(self):
        if not self._driver:
            self._make_driver()
        return self

    def get_page(self, url, *args, **kwargs):
        return self._driver.get(url, *args, **kwargs)

    def get_cfpage(self, url, *args, **kwargs):
        page = self.get_page(url, *args, **kwargs)

        WebDriverWait(self._driver, 15).until_not(
            EC.presence_of_element_located(
                (By.CSS_SELECTOR, 'p[data-translate=resolve_captcha_network]')
            )
        )

        return page

    @property
    def source(self):
        return self._driver.page_source

    def __del__(self):
        if self._driver:
            self.exit()

    def exit(self):
        self._driver.quit()



# Models
from mext.models.chapter import Chapter
from mext.models.cover import Cover
from mext.models.genre import Genre
from mext.models.manga import Manga
from mext.models.page import Page
from mext.models.person import Person
from mext.models.link import Link
from mext.models.group import Group
from mext.models.tag import Tag


__all__ = [
    Chapter,
    Cover,
    Genre,
    Manga,
    Page,
    Person,
    Link,
    Group,
    Tag
]



import re

from mext.provider import Provider

from .p_asurascans_com import AsuraScansCom
from .p_mangadex_org import MangadexOrg
from .p_reaperscans_com import ReaperScansCom


ALL_PROVIDERS = [
    {
        "name": "AsuraScans",
        "regex": r".*(asura\.gg|asurascans\.com).*",
        "class": AsuraScansCom
    },
    {
        "name": "ReaperScans",
        "regex": r".*(reaperscans\.com).*",
        "class": ReaperScansCom
    },
    {
        "name": "MangaDex",
        "regex": r".*(mangadex\.org).*",
        "class": MangadexOrg
    }
]


def get_provider_instance(name=None, netloc=None) -> Provider:
    for provider_info in ALL_PROVIDERS:
        if (name and provider_info['name'] == name) or \
                (netloc and re.search(provider_info['regex'], netloc)):
            ProviderClass = provider_info['class']
            return ProviderClass(
                name=provider_info['name'],
                siteUrl=netloc,
            )
    raise ValueError("No provider for provided netloc {}".format(netloc))


__all__ = [
    get_provider_instance
]



import io

from .model import Model

class Cover(Model):
    """Represents a Cover."""
    __slots__ = (
        "id", "description", "volume", "file_bytes", "image_link", "url",
        "provider", "instance"
    )

    def __init__(self, provider):
        super().__init__(provider)

        self.id: str = ""
        self.description: str = ""
        self.volume: float = float(0)
        self.filename: str = ""
        self.file_bytes: io.BytesIO = io.BytesIO()
        self.image_link: str = ""
        self.url: str = ""
        self.provider = provider
        self.instance = self

    def __str__(self) -> str:
        return self.image_link

    def __repr__(self) -> str:
        return str(self)

    def get_url(self) -> str:
        return self.image_link or ""



from typing import List, Dict
from datetime import datetime

from mext import models

from .model import Model


class Manga(Model):
    """Represents a Manga."""
    __slots__ = (
        "id", "title", "alts", "description", "links", "language", "comic_type", "status",
        "year", "rating", "followers", "genres", "tags", "authors", "artists", "current_cover",
        "all_covers", "banner_picture", "content_rating", "first_chapter", "last_chapter", "chapter_list",
        "url", "created_at", "updated_at", "provider", "instance"
    )

    def __init__(self, provider):
        super().__init__(provider)
        
        self.id: str = ""
        self.title: str = ""
        self.alts: List[str] = []
        self.description: str = ""
        self.links: Dict = []
        self.language: str = None
        self.comic_type: str = ""
        self.status: str = None
        self.year: int = datetime.now().year
        self.rating: float = float(0)
        self.followers: int = int(0)
        self.genres: List[models.Genre] = []
        self.tags: List[models.Tag] = []
        self.authors: List[models.Person] = []
        self.artists: List[models.Person] = []
        self.current_cover: models.Cover = None
        self.all_covers: List[models.Cover] = []
        self.banner_picture: str = ""
        self.content_rating: str = ""
        self.first_chapter: models.Chapter = None
        self.last_chapter: models.Chapter = None
        self.chapter_list: List[models.Chapter] = []
        self.url: str = ""
        self.created_at: datetime = None
        self.updated_at: datetime = None
        self.provider = provider
        self.instance = self

    def __str__(self) -> str:
        return self.title

    def __repr__(self) -> str:
        return str(self)
